@@ -14,6 +14,8 @@
 
 package cmd
 
+//go:generate go run gen_experiments_help.go
+
 import (
 	"bufio"
 	"fmt"
@@ -125,6 +127,7 @@ var helpTopics = []*cobra.Command{
 	commandsHelp,
 	embedHelp,
 	environmentHelp,
+	experimentsHelp,
 	filetypeHelp,
 	flagsHelp,
 	injectHelp,
@@ -161,6 +164,10 @@ to a registry location by consulting cue.mod/module.cue and the
 registry configuration (see "cue help modules" and "cue help
 registryconfig" for more details).
 
+A package is loaded as a package instance: the unification of all CUE
+files with the same package name found in the package's directory and
+every ancestor directory leading up to the module root.
+
 An import may contain a ":name" suffix to indicate a specific
 package to load within a given path. It is necessary to specify this
 when there is more than one package at the path.
@@ -244,6 +251,12 @@ which inserts one field per file matched:
 
 Note that "**" glob patterns are not supported at this time.
 
+By default, a glob pattern that matches no files results in an error.
+To allow empty results, use the "allowEmptyGlob" option:
+
+	c: _ @embed(glob=optional/*.json, allowEmptyGlob)
+	c: [string]: {...}
+
 If the file extension in "file" or "glob" does not imply a file type,
 it must be specified with the "type" encoding as shown above.
 See the "filetypes" help topic for more.
@@ -295,16 +308,8 @@ If an environment variable is unset or empty, sensible default setting is used.
 		See "cue help registryconfig" for details.
 
 	CUE_EXPERIMENT
-		Comma-separated list of experiment flags to enable or disable:
-
-		evalv3 (default true)
-			Enable the new CUE evaluator, addressing performance issues
-			and bringing a better disjunction algorithm.
-		cmdreferencepkg (default true)
-			Require referencing imported tool packages to declare "cue cmd" tasks.
-		keepvalidators (default true)
-			Prevent validators from simplifying into concrete values.
-			See: https://cuelang.org/discussion/3775
+		Comma-separated list of experiment flags to enable or disable.
+		See "cue help experiments" for details.
 
 	CUE_DEBUG
 		Comma-separated list of debug flags to enable or disable, such as:
