@@ -12,7 +12,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-package cmd
+package cmd_test
 
 import (
 	"bufio"
@@ -46,10 +46,12 @@ import (
 	"golang.org/x/oauth2"
 	"golang.org/x/tools/txtar"
 
+	cuecmd "cuelang.org/go/cmd/cue/cmd"
 	"cuelang.org/go/cue"
 	"cuelang.org/go/cue/cuecontext"
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/cue/parser"
+	"cuelang.org/go/internal/cuedebug"
 	"cuelang.org/go/internal/cuetest"
 	"cuelang.org/go/internal/cueversion"
 	"cuelang.org/go/internal/mod/semver"
@@ -82,8 +84,7 @@ func TestLatest(t *testing.T) {
 				if !strings.HasSuffix(f.Name, ".cue") || path.Base(f.Name) == "invalid.cue" {
 					return
 				}
-				v := parser.FromVersion(parser.Latest)
-				_, err := parser.ParseFile(f.Name, f.Data, v)
+				_, err := parser.ParseFile(f.Name, f.Data)
 				if err != nil {
 					w := &bytes.Buffer{}
 					fmt.Fprintf(w, "\n%s:\n", fullpath)
@@ -354,6 +355,10 @@ func TestScript(t *testing.T) {
 // Usage Comment out t.Skip() and set file to test.
 func TestX(t *testing.T) {
 	t.Skip()
+	// adt.OpenGraphs = true
+	// adt.DebugDeps = true
+	cuedebug.Init()
+	cuedebug.Flags.LogEval = 1
 	const path = "./testdata/script/eval_e.txtar"
 
 	check := func(err error) {
@@ -374,10 +379,7 @@ func TestX(t *testing.T) {
 		check(os.WriteFile(name, f.Data, 0666))
 	}
 
-	cwd, err := os.Getwd()
-	check(err)
-	defer func() { _ = os.Chdir(cwd) }()
-	_ = os.Chdir(tmpdir)
+	t.Chdir(tmpdir)
 
 	for s := bufio.NewScanner(bytes.NewReader(a.Comment)); s.Scan(); {
 		cmd := s.Text()
@@ -390,7 +392,7 @@ func TestX(t *testing.T) {
 		args, err := shlex.Split(cmd)
 		check(err)
 
-		c, _ := New(args[1:])
+		c, _ := cuecmd.New(args[1:])
 		b := &bytes.Buffer{}
 		c.SetOutput(b)
 		err = c.Run(context.Background())
@@ -409,15 +411,15 @@ func TestMain(m *testing.M) {
 		}
 	}
 	testscript.Main(m, map[string]func(){
-		"cue": func() { os.Exit(Main()) },
+		"cue": func() { os.Exit(cuecmd.Main()) },
 		// Until https://github.com/rogpeppe/go-internal/issues/93 is fixed,
 		// or we have some other way to use "exec" without caring about success,
 		// this is an easy way for us to mimic `? exec cue`.
-		"cue_exitzero": func() { Main() },
+		"cue_exitzero": func() { cuecmd.Main() },
 		"cue_stdinpipe": func() {
 			cwd, _ := os.Getwd()
 			if err := mainStdinPipe(); err != nil {
-				if err != ErrPrintedError { // print errors like Main
+				if err != cuecmd.ErrPrintedError { // print errors like Main
 					errors.Print(os.Stderr, err, &errors.Config{
 						Cwd:     cwd,
 						ToSlash: testing.Testing(),
@@ -488,7 +490,7 @@ func tsExpand(ts *testscript.TestScript, s string) string {
 func mainStdinPipe() error {
 	// Like Main, but sets stdin to a pipe,
 	// to emulate stdin reads like a terminal.
-	cmd, _ := New(os.Args[1:])
+	cmd, _ := cuecmd.New(os.Args[1:])
 	pr, pw, err := os.Pipe()
 	if err != nil {
 		return err
@@ -552,6 +554,11 @@ func testCmd() {
 				fmt.Println(msg)
 			}
 		}
+	case "expand":
+		// Expands env vars like $FOO and ${BAR} in arguments, and prints them out.
+		for _, arg := range args {
+			fmt.Println(os.ExpandEnv(arg))
+		}
 	default:
 		check(fmt.Errorf("unknown command: %q\n", cmd))
 	}
@@ -595,6 +602,8 @@ func newMockRegistryOauth(mode string) *httptest.Server {
 		switch mode {
 		case "device-code-expired":
 			writeJSON(w, http.StatusBadRequest, tokenError{ErrorCode: tokenErrorCodeExpired})
+		case "pending-forever":
+			writeJSON(w, http.StatusBadRequest, tokenError{ErrorCode: tokenErrorCodePending})
 		case "pending-success":
 			count := tokenRequestCounter.Add(1)
 			if count == 1 {
