@@ -15,7 +15,6 @@
 package cue_test
 
 import (
-	"bytes"
 	"fmt"
 	"io"
 	"math"
@@ -25,6 +24,7 @@ import (
 	"strings"
 	"testing"
 
+	"github.com/go-quicktest/qt"
 	"github.com/google/go-cmp/cmp"
 
 	"cuelang.org/go/cue"
@@ -861,7 +861,7 @@ func TestFields(t *testing.T) {
 		res:   "{reg:4,}",
 	}, {
 		value: `{a:1,b:2,c:int}`,
-		err:   "cannot convert incomplete value",
+		res:   "{a:1,b:2,c:int,}",
 	}, {
 		value: `
 			step1: {}
@@ -870,7 +870,7 @@ func TestFields(t *testing.T) {
 			   step3: {prefix: step2.value}
 			}
 			_hidden: 3`,
-		res: `{step1:{},step2:{"prefix":3},}`,
+		res: `{step1:,step2:prefix: 3,}`,
 	}, {
 		opts: []cue.Option{cue.Final()},
 		value: `
@@ -904,9 +904,37 @@ func TestFields(t *testing.T) {
 		value: `a: x, x: y, y: b: 1`,
 		path:  "a",
 		res:   `{b:1,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `[=~"^a"]: 1`,
+		res:   `{[=~"^a"]:1,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `[=~"^a"]: 1, [string]: 2`,
+		res:   `{[=~"^a"]:1,[string]:2,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `[=~"^a"]: >1, a: 2`,
+		res:   `{[=~"^a"]:>1,a:2,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{[_]: 1}`,
+		res:   `{[_]:1,}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{...}`,
+		res:   `{}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{...}`,
+		res:   `{}`,
+	}, {
+		opts:  []cue.Option{cue.Patterns(true)},
+		value: `{[1, 2, ...int], [string]: int}`,
+		res:   `{[>=2]:int,[string]:int,0:1,1:2,}`,
 	}}
 	for _, tc := range testCases {
-		cuetdtest.DevOnlyMatrix.Run(t, tc.value, func(t *testing.T, m *cuetdtest.M) {
+		cuetdtest.SmallMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
 			f := getValue(m, tc.value)
 
 			obj := f.LookupPath(cue.ParsePath(tc.path))
@@ -916,11 +944,23 @@ func TestFields(t *testing.T) {
 
 			buf := []byte{'{'}
 			for iter.Next() {
-				buf = append(buf, iter.Selector().String()...)
+				sel := iter.Selector()
+				if sel.ConstraintType() == cue.PatternConstraint {
+					qt.Check(t, qt.Equals(
+						sel.String(),
+						fmt.Sprintf("[%#v]", sel.Pattern()),
+					))
+					wantType := cue.StringLabel | cue.PatternConstraint
+					if obj.Kind() == cue.ListKind {
+						wantType = cue.IndexLabel | cue.PatternConstraint
+					}
+					qt.Check(t, qt.Equals(sel.Type(), wantType))
+				}
+				buf = append(buf, sel.String()...)
 				buf = append(buf, ':')
-				b, err := iter.Value().MarshalJSON()
-				checkFatal(t, err, tc.err, "Obj.At")
-				buf = append(buf, b...)
+				v := iter.Value()
+				checkFatal(t, v.Err(), tc.err, "Obj.At")
+				buf = fmt.Appendf(buf, "%#v", v)
 				buf = append(buf, ',')
 			}
 			buf = append(buf, '}')
@@ -930,14 +970,20 @@ func TestFields(t *testing.T) {
 
 			iter, _ = obj.Fields(tc.opts...)
 			for iter.Next() {
-				want, err := iter.Value().MarshalJSON()
-				checkFatal(t, err, tc.err, "Obj.At2")
+				v := iter.Value()
+				checkFatal(t, v.Err(), tc.err, "Obj.At2")
+				want := fmt.Sprintf("%#v", v)
 
-				got, err := obj.LookupPath(cue.MakePath(iter.Selector())).MarshalJSON()
-				checkFatal(t, err, tc.err, "Obj.At2")
-
-				if !bytes.Equal(got, want) {
-					t.Errorf("Lookup: got %q; want %q", got, want)
+				got := obj.LookupPath(cue.MakePath(iter.Selector()))
+				if iter.FieldType().ConstraintType() == cue.PatternConstraint {
+					// Can't look up iterated pattern constraints.
+					qt.Assert(t, qt.ErrorMatches(
+						got.Err(),
+						`.*cannot look up pattern constraints.*`,
+					))
+				} else {
+					checkFatal(t, err, tc.err, "Obj.At2")
+					qt.Assert(t, qt.Equals(fmt.Sprintf("%#v", got), want))
 				}
 			}
 			v := obj.LookupPath(cue.MakePath(cue.Str("non-existing")))
@@ -2561,6 +2607,25 @@ func TestIssue3826(t *testing.T) {
 	})
 }
 
+// TestIssue4093 tests that a value that's a *adt.NodeLink
+// under the hood can be formatted OK.
+func TestIssue4093(t *testing.T) {
+	cuetdtest.FullMatrix.Run(t, "test", func(t *testing.T, m *cuetdtest.M) {
+		v := m.CueContext().CompileString(`x: #X, #X: int`)
+		qt.Assert(t, qt.IsNil(v.Err()))
+		v = v.LookupPath(cue.MakePath(cue.Str("x")))
+		op, args := v.Expr()
+		qt.Assert(t, qt.Equals(op, cue.SelectorOp))
+		qt.Assert(t, qt.HasLen(args, 2))
+		// Note: we see the entire value because the #X is represented
+		// internally as a selector operation topLevelValue.#X
+		// and we're looking at the first argument to that operation.
+		// It's not clear why we see `x: int` not `x: #X`.
+		s := fmt.Sprintf("%#v", args[0])
+		qt.Assert(t, qt.Equals(s, "x:  int\n#X: int"))
+	})
+}
+
 func TestEquals(t *testing.T) {
 	testCases := []struct {
 		a, b string
@@ -2973,7 +3038,6 @@ func TestValueDoc(t *testing.T) {
 			val  cue.Value
 			path string
 			doc  string
-			skip bool
 		}{{
 			val:  v1,
 			path: "foos",
@@ -3005,12 +3069,10 @@ field1 is an int.
 		}, {
 			val:  v1,
 			path: "baz field1",
-			doc: `comment from bar on field 1
+			doc: `comment from baz on field 1
 
-comment from baz on field 1
+comment from bar on field 1
 `,
-			// New evaluaotor orders the comments differently (arguably better).
-			skip: true,
 		}, {
 			val:  v1,
 			path: "baz field2",
@@ -3029,9 +3091,6 @@ Another Foo.
 `,
 		}}
 		for _, tc := range testCases {
-			if tc.skip {
-				m.TODO_V3(t) // P2: reordering
-			}
 			t.Run("field:"+tc.path, func(t *testing.T) {
 				v := tc.val.Lookup(strings.Split(tc.path, " ")...)
 				doc := docStr(v.Doc())
@@ -3189,7 +3248,7 @@ func TestMarshalJSON(t *testing.T) {
 	}, {
 		// Issue #326
 		value: `x: "\(string)": "v"`,
-		err:   `x: invalid interpolation`,
+		err:   `x: key value of dynamic field must be concrete, found _|_(x: invalid interpolation: non-concrete value string (type string)) (and 1 more errors)`,
 	}, {
 		// Issue #326
 		value: `x: "\(bool)": "v"`,
@@ -3232,8 +3291,6 @@ func TestMarshalJSON(t *testing.T) {
 	}}
 	for i, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, fmt.Sprintf("%d/%v", i, tc.value), func(t *testing.T, m *cuetdtest.M) {
-			m.TODO_V3(t) // P1: wrong output
-
 			val := getValue(m, tc.value)
 			b, err := val.MarshalJSON()
 			checkFatal(t, err, tc.err, "init")
@@ -3496,7 +3553,6 @@ func TestPos(t *testing.T) {
 	testCases := []struct {
 		value string
 		pos   string
-		skip  bool
 	}{{
 		value: `
 a: string
@@ -3506,11 +3562,7 @@ a: "foo"`,
 		value: `
 a: x: string
 a: x: "x"`,
-		pos: "2:4",
-
-		// the position of the new evaluator is also correct, and actually
-		// better.
-		skip: true,
+		pos: "3:4",
 	}, {
 		// Prefer struct conjuncts with actual fields.
 		value: `
@@ -3530,10 +3582,6 @@ a: x: y: z: "x"`,
 	}}
 	for _, tc := range testCases {
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
-			if tc.skip {
-				m.TODO_V3(t) // P3: different position
-			}
-
 			c := m.CueContext()
 			v := c.CompileString(tc.value)
 			v = v.LookupPath(cue.ParsePath("a"))
@@ -3550,7 +3598,6 @@ func TestPathCorrection(t *testing.T) {
 		input  string
 		lookup func(i cue.Value) cue.Value
 		want   string
-		skip   bool
 	}{{
 		input: `
 			a: b: {
@@ -3788,9 +3835,6 @@ func TestPathCorrection(t *testing.T) {
 		},
 	}}
 	for _, tc := range testCases {
-		if tc.skip {
-			continue
-		}
 		cuetdtest.FullMatrix.Run(t, "", func(t *testing.T, m *cuetdtest.M) {
 			ctx := m.CueContext()
 
@@ -3992,6 +4036,19 @@ func TestExpr(t *testing.T) {
 	}, {
 		input: "v: 2 mod 5",
 		want:  "mod(2 5)",
+	}, {
+		input: "@experiment(explicitopen)\n v: #Y..., #Y: {b: 2}",
+		want:  `...(.(〈〉 "#Y"))`,
+	}, {
+		input: "@experiment(explicitopen)\n v: {#Y..., a: 1}, #Y: {b: 2}",
+		want:  `&(...(.(〈〉 "#Y")) {a:1})`,
+	}, {
+		input: "@experiment(explicitopen)\n v: {#Y, a: 1}, #Y: {b: 2}",
+		want:  `&(.(〈〉 "#Y") {a:1})`,
+	}, {
+		// Note: before explicitopen there is no way to detect embeddings.
+		input: "v: {#Y, a: 1}, #Y: {b: 2}",
+		want:  `&(.(〈〉 "#Y") {a:1})`,
 	}, {
 		input: "v: a.b, a: b: 4",
 		want:  `.(.(〈〉 "a") "b")`,
@@ -4097,3 +4154,300 @@ func compactRawStr(v cue.Value) string {
 	cfg := &debug.Config{Compact: true, Raw: true}
 	return debug.NodeString(ctx, cue.ValueVertex(v), cfg)
 }
+
+func TestIsClosed(t *testing.T) {
+	testCases := []struct {
+		desc                string
+		input               string
+		path                string
+		wantClosed          bool
+		wantClosedRecursive bool
+	}{
+		{
+			desc:                "open struct",
+			input:               `x: {a: 1}`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "struct with close builtin",
+			input:               `x: close({a: 1})`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "definition",
+			input:               `#X: {a: 1}`,
+			path:                "#X",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "field in definition",
+			input:               `#X: {a: {b: 1}}`,
+			path:                "#X.a",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: definition reference",
+			input:               `#X: {a: 1}, y: #X`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: definition nested field",
+			input:               `#X: {a: {b: 1}}, y: #X`,
+			path:                "y.a",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: mixed open and closed",
+			input:               `#X: {a: 1}, y: {b: 2, c: #X}`,
+			path:                "y.c",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "structure sharing: open struct in closed",
+			input:               `#X: {a: open}, open: {b: 1}, y: #X.a`,
+			path:                "y",
+			wantClosed:          true, // Inherits closed status from definition
+			wantClosedRecursive: true, // Inherits closed status from definition
+		},
+		{
+			desc:                "disjunction with close",
+			input:               `x: close({a: 1}) | close({b: 2})`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "disjunction with close and elimination",
+			input:               `x: close({a: 1}) | close({b: 2}), x: a: 1`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "disjunction with definition",
+			input:               `#A: {a: 1}, #B: {b: 2}, x: #A | #B`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false, // Disjunction itself is not closed
+		},
+		{
+			desc:                "disjunction with mixed closedness",
+			input:               `#A: {a: 1}, x: #A | {b: 2}`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "disjunction resolved to closed",
+			input:               `#A: {a: 1}, #B: {b: 2}, x: (#A | #B) & {a: 1}`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "nested definition in open struct",
+			input:               `x: {#D: {a: 1}}`,
+			path:                "x.#D",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "empty struct with close",
+			input:               `x: close({})`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "empty definition",
+			input:               `#X: {}`,
+			path:                "#X",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "close with pattern constraint",
+			input:               `x: close({[string]: int})`,
+			path:                "x",
+			wantClosed:          true,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "definition with pattern constraint",
+			input:               `#X: {[string]: int}`,
+			path:                "#X",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 1",
+			input:               `#A: {a: 1}, #B: {b: #A}, y: #B.b`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 2",
+			input:               `#A: {a: 1}, x: {b: #A}, y: x.b`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 3",
+			input:               `#A: {a: {c: 1}}, x: {b: #A}, y: x.b.a`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "multiple levels of structure sharing 4",
+			input:               `a: {a: c: 1}, x: {#b: a}, y: x.#b.a`,
+			path:                "y",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "scalar value (non-struct)",
+			input:               `x: 42`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "list value",
+			input:               `x: [1, 2, 3]`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "unresolved disjunction with all closed options",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				x: #A | #B
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "unresolved disjunction with mixed closedness",
+			input: `
+				#A: {a: 1}
+				x: #A | {b: 2}
+			`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "disjunction with close() builtin",
+			input: `
+				x: close({a: 1}) | close({b: 2})
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "default disjunction with closed values",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				x: *#A | #B
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "nested disjunction in definition",
+			input: `
+				x: #X
+				#X: {
+					field: close({a: 1}) | close({b: 2})
+				}
+			`,
+			path:                "x",
+			wantClosed:          true, // Definition is closed
+			wantClosedRecursive: true, // Definition is recursively closed
+		},
+		{
+			desc: "triple disjunction all closed",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				#C: {c: 3}
+				x: #A | #B | #C
+			`,
+			path:                "x",
+			wantClosed:          false, // Disjunction itself is not closed
+			wantClosedRecursive: false,
+		},
+		{
+			desc: "triple disjunction with one open",
+			input: `
+				#A: {a: 1}
+				#B: {b: 2}
+				x: #A | #B | {c: 3}
+			`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "definition with disjunction of scalar and struct",
+			input:               `#A: 1 | {a!: int}`,
+			path:                "#A",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+		{
+			desc:                "closed list with definition elements",
+			input:               `#A: {a: 1}, x: [#A, #A]`,
+			path:                "x",
+			wantClosed:          false,
+			wantClosedRecursive: false,
+		},
+		{
+			desc:                "list element from definition",
+			input:               `#A: {a: 1}, x: [#A, #A]`,
+			path:                "x[0]",
+			wantClosed:          true,
+			wantClosedRecursive: true,
+		},
+	}
+
+	for _, tc := range testCases {
+		t.Run(tc.desc, func(t *testing.T) {
+			cuetdtest.FullMatrix.Run(t, tc.input, func(t *testing.T, m *cuetdtest.M) {
+				v := getValue(m, tc.input)
+
+				v = v.LookupPath(cue.ParsePath(tc.path))
+
+				if err := v.Err(); err != nil {
+					t.Fatalf("unexpected error: %v", err)
+				}
+
+				qt.Check(t, qt.Equals(v.IsClosed(), tc.wantClosed))
+				qt.Check(t, qt.Equals(v.IsClosedRecursively(), tc.wantClosedRecursive))
+
+				if v.IsClosedRecursively() {
+					qt.Check(t, qt.IsTrue(v.IsClosed()))
+				}
+			})
+		})
+	}
+}
