@@ -16,12 +16,13 @@ package adt
 
 import (
 	"fmt"
+	"iter"
 	"slices"
 
 	"cuelang.org/go/cue/ast"
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/cue/token"
-	"cuelang.org/go/internal"
+	"cuelang.org/go/internal/iterutil"
 )
 
 // TODO: unanswered questions about structural cycles:
@@ -111,38 +112,6 @@ func (e *Environment) up(ctx *OpContext, count int32) *Environment {
 	return e
 }
 
-type ID int32
-
-// evalCached is used to look up dynamic field pattern constraint expressions.
-func (e *Environment) evalCached(c *OpContext, x Expr) Value {
-	if v, ok := x.(Value); ok {
-		return v
-	}
-	key := cacheKey{x, nil}
-	v, ok := e.cache[key]
-	if !ok {
-		if e.cache == nil {
-			e.cache = map[cacheKey]Value{}
-		}
-		env, src := c.e, c.src
-		c.e, c.src = e, x.Source()
-		// Save and restore errors to ensure that only relevant errors are
-		// associated with the cash.
-		err := c.errs
-		v = c.evalState(x, combinedFlags{
-			status:    partial,
-			condition: allKnown,
-			mode:      yield,
-		}) // TODO: should this be finalized?
-		c.e, c.src = env, src
-		c.errs = err
-		if b, ok := v.(*Bottom); !ok || !b.IsIncomplete() {
-			e.cache[key] = v
-		}
-	}
-	return v
-}
-
 // A Vertex is a node in the value tree. It may be a leaf or internal node.
 // It may have arcs to represent elements of a fully evaluated struct or list.
 //
@@ -174,11 +143,6 @@ type Vertex struct {
 	// status indicates the evaluation progress of this vertex.
 	status vertexStatus
 
-	// hasAllConjuncts indicates that the set of conjuncts is complete.
-	// This is the case if the conjuncts of all its ancestors have been
-	// processed.
-	hasAllConjuncts bool
-
 	// isData indicates that this Vertex is to be interpreted as data: pattern
 	// and additional constraints, as well as optional fields, should be
 	// ignored.
@@ -193,6 +157,10 @@ type Vertex struct {
 	// level only. This supports the close builtin.
 	ClosedNonRecursive bool
 
+	// Opened is set when a node that is opened with @experiment(explicitopen)
+	// is structure shared. This will override any of the above booleans.
+	OpenedShared bool
+
 	// HasEllipsis indicates that this Vertex is open by means of an ellipsis.
 	// TODO: combine this field with Closed once we removed the old evaluator.
 	HasEllipsis bool
@@ -202,11 +170,6 @@ type Vertex struct {
 	// the per-Environment value cache.
 	MultiLet bool
 
-	// After this is set, no more arcs may be added during evaluation. This is
-	// set, for instance, after a Vertex is used as a source for comprehensions,
-	// or any other operation that relies on the set of arcs being constant.
-	LockArcs bool
-
 	// IsDynamic signifies whether this struct is computed as part of an
 	// expression and not part of the static evaluation tree.
 	// Used for cycle detection.
@@ -230,9 +193,6 @@ type Vertex struct {
 	// the for source of comprehensions and let fields or let clauses.
 	anonymous bool
 
-	// hasPendingArc is set if this Vertex has a void arc (e.g. for comprehensions)
-	hasPendingArc bool
-
 	// IsDisjunct indicates this Vertex is a disjunct resulting from a
 	// disjunction evaluation.
 	IsDisjunct bool
@@ -242,18 +202,9 @@ type Vertex struct {
 	// The debug printer, for instance, takes extra care not to print in a loop.
 	IsShared bool
 
-	// IsCyclic is true if a node is cyclic, for instance if its value is
-	// a cyclic reference to a shared node or if the value is a conjunction
-	// of which at least one value is cyclic (not yet supported).
-	IsCyclic bool
-
 	// ArcType indicates the level of optionality of this arc.
 	ArcType ArcType
 
-	// cyclicReferences is a linked list of internal references pointing to this
-	// Vertex. This is used to shorten the path of some structural cycles.
-	cyclicReferences *RefNode
-
 	// BaseValue is the value associated with this vertex. For lists and structs
 	// this is a sentinel value indicating its kind.
 	BaseValue BaseValue
@@ -280,8 +231,8 @@ type Vertex struct {
 	// the final value of this Vertex.
 	//
 	// TODO: all access to Conjuncts should go through functions like
-	// VisitLeafConjuncts and VisitAllConjuncts. We should probably make this
-	// an unexported field.
+	// [Vertex.LeafConjuncts] and [Vertex.AllConjuncts].
+	// We should probably make this an unexported field.
 	Conjuncts ConjunctGroup
 
 	// Structs is a slice of struct literals that contributed to this value.
@@ -308,6 +259,9 @@ func equalDeref(a, b *Vertex) bool {
 // newInlineVertex creates a Vertex that is needed for computation, but for
 // which there is no CUE path defined from the root Vertex.
 func (ctx *OpContext) newInlineVertex(parent *Vertex, v BaseValue, a ...Conjunct) *Vertex {
+	// TODO: parent is an unused parameter here. Setting [Vertex.Parent] to it
+	// improves paths in a bunch of errors, fixing regressions compared to evalv2.
+	// However, it also breaks a few tests. Perhaps try with evalv4.
 	n := &Vertex{
 		BaseValue: v,
 		IsDynamic: true,
@@ -318,9 +272,6 @@ func (ctx *OpContext) newInlineVertex(parent *Vertex, v BaseValue, a ...Conjunct
 		state := ctx.freeScope[len(ctx.freeScope)-1]
 		state.toFree = append(state.toFree, n)
 	}
-	if !ctx.isDevVersion() {
-		n.Parent = parent
-	}
 	if ctx.inDetached > 0 {
 		n.anonymous = true
 	}
@@ -350,7 +301,7 @@ func (v *Vertex) updateArcType(t ArcType) {
 			return
 		}
 	}
-	if v.Parent != nil && v.Parent.ArcType == ArcPending && v.Parent.state != nil && v.Parent.state.ctx.isDevVersion() {
+	if v.Parent != nil && v.Parent.ArcType == ArcPending && v.Parent.state != nil {
 		// TODO: check that state is always non-nil.
 		v.Parent.state.unshare()
 	}
@@ -420,6 +371,8 @@ func (v *Vertex) MayAttach() bool {
 	return !v.Label.IsLet() && !v.anonymous
 }
 
+//go:generate go tool stringer -type=ArcType -trimprefix=Arc
+
 type ArcType uint8
 
 const (
@@ -454,29 +407,6 @@ const (
 	// We could also define types for required fields and potentially lets.
 )
 
-func (a ArcType) String() string {
-	switch a {
-	case ArcMember:
-		return "Member"
-	case ArcOptional:
-		return "Optional"
-	case ArcRequired:
-		return "Required"
-	case ArcPending:
-		return "Pending"
-	case ArcNotPresent:
-		return "NotPresent"
-	}
-	return fmt.Sprintf("ArcType(%d)", a)
-}
-
-// definitelyExists reports whether an arc is a constraint or member arc.
-// TODO: we should check that users of this call ensure there are no
-// ArcPendings.
-func (v *Vertex) definitelyExists() bool {
-	return v.ArcType < ArcPending
-}
-
 // ConstraintFromToken converts a given AST constraint token to the
 // corresponding ArcType.
 func ConstraintFromToken(t token.Token) ArcType {
@@ -536,24 +466,12 @@ type StructInfo struct {
 	// or nil if this is a root structure.
 	// Embed   *StructInfo
 	// Context *RefInfo // the location from which this struct originates.
-	Disable bool
-
-	Embedding bool
-}
-
-// TODO(perf): this could be much more aggressive for eliminating structs that
-// are immaterial for closing.
-func (s *StructInfo) useForAccept() bool {
-	if c := s.closeInfo; c != nil {
-		return !c.noCheck
-	}
-	return true
 }
 
 // vertexStatus indicates the evaluation progress of a Vertex.
 type vertexStatus int8
 
-//go:generate go run golang.org/x/tools/cmd/stringer -type=vertexStatus
+//go:generate go tool stringer -type=vertexStatus
 
 const (
 	// unprocessed indicates a Vertex has not been processed before.
@@ -575,16 +493,6 @@ const (
 	// but without recursively processing arcs.
 	conjuncts
 
-	// evaluatingArcs indicates that the arcs of the Vertex are currently being
-	// evaluated. If this is encountered it indicates a structural cycle.
-	// Value does not have to be nil
-	evaluatingArcs
-
-	// TODO: introduce a "frozen" state. Right now a node may marked and used
-	// as finalized, before all tasks have completed. We should introduce a
-	// frozen state that simply checks that all remaining tasks are idempotent
-	// and errors if they are not.
-
 	// finalized means that this node is fully evaluated and that the results
 	// are save to use without further consideration.
 	finalized
@@ -645,29 +553,32 @@ func (v *Vertex) updateStatus(s vertexStatus) {
 // received all their conjuncts as well, after which this node will have been
 // notified of these conjuncts.
 func (v *Vertex) setParentDone() {
-	v.hasAllConjuncts = true
 	// Could set "Conjuncts" flag of arc at this point.
-	if n := v.state; n != nil && len(n.conjuncts) == n.conjunctsPos {
+	if n := v.state; n != nil {
 		for _, a := range v.Arcs {
 			a.setParentDone()
 		}
 	}
 }
 
-// VisitLeafConjuncts visits all conjuncts that are leafs of the ConjunctGroup tree.
-func (v *Vertex) VisitLeafConjuncts(f func(Conjunct) bool) {
-	VisitConjuncts(v.Conjuncts, f)
+// LeafConjuncts iterates over all conjuncts that are leaves of the [ConjunctGroup] tree.
+func (v *Vertex) LeafConjuncts() iter.Seq[Conjunct] {
+	return func(yield func(Conjunct) bool) {
+		_ = iterConjuncts(v.Conjuncts, yield)
+	}
 }
 
-func VisitConjuncts(a []Conjunct, f func(Conjunct) bool) bool {
+func iterConjuncts(a []Conjunct, yield func(Conjunct) bool) bool {
+	// TODO: note that this is iterAllConjuncts but without yielding ConjunctGroups.
+	// Can we reuse the code in a simple enough way?
 	for _, c := range a {
 		switch x := c.x.(type) {
 		case *ConjunctGroup:
-			if !VisitConjuncts(*x, f) {
+			if !iterConjuncts(*x, yield) {
 				return false
 			}
 		default:
-			if !f(c) {
+			if !yield(c) {
 				return false
 			}
 		}
@@ -675,22 +586,39 @@ func VisitConjuncts(a []Conjunct, f func(Conjunct) bool) bool {
 	return true
 }
 
-// VisitAllConjuncts visits all conjuncts of v, including ConjunctGroups.
-// Note that ConjunctGroups do not have an Environment associated with them.
-func (v *Vertex) VisitAllConjuncts(f func(c Conjunct, isLeaf bool)) {
-	visitAllConjuncts(v.Conjuncts, f)
+// ConjunctsSeq iterates over all conjuncts that are leafs in the list of trees given.
+func ConjunctsSeq(a []Conjunct) iter.Seq[Conjunct] {
+	return func(yield func(Conjunct) bool) {
+		_ = iterConjuncts(a, yield)
+	}
 }
 
-func visitAllConjuncts(a []Conjunct, f func(c Conjunct, isLeaf bool)) {
+// AllConjuncts iterates through all conjuncts of v, including [ConjunctGroup]s.
+// Note that ConjunctGroups do not have an Environment associated with them.
+// The boolean reports whether the conjunct is a leaf.
+func (v *Vertex) AllConjuncts() iter.Seq2[Conjunct, bool] {
+	return func(yield func(Conjunct, bool) bool) {
+		_ = iterAllConjuncts(v.Conjuncts, yield)
+	}
+}
+
+func iterAllConjuncts(a []Conjunct, yield func(c Conjunct, isLeaf bool) bool) bool {
 	for _, c := range a {
 		switch x := c.x.(type) {
 		case *ConjunctGroup:
-			f(c, false)
-			visitAllConjuncts(*x, f)
+			if !yield(c, false) {
+				return false
+			}
+			if !iterAllConjuncts(*x, yield) {
+				return false
+			}
 		default:
-			f(c, true)
+			if !yield(c, true) {
+				return false
+			}
 		}
 	}
+	return true
 }
 
 // HasConjuncts reports whether v has any conjuncts.
@@ -707,14 +635,11 @@ func (v *Vertex) SingleConjunct() (c Conjunct, count int) {
 	if v == nil {
 		return c, 0
 	}
-	v.VisitLeafConjuncts(func(x Conjunct) bool {
-		c = x
+	for c = range v.LeafConjuncts() {
 		if count++; count > 1 {
-			return false
+			break
 		}
-		return true
-	})
-
+	}
 	return c, count
 }
 
@@ -827,12 +752,6 @@ func (v *Vertex) toDataAllRec(ctx *OpContext, processed map[*Vertex]*Vertex) *Ve
 	w.ClosedRecursive = false
 	w.ClosedNonRecursive = false
 
-	// TODO(perf): this is not strictly necessary for evaluation, but it can
-	// hurt performance greatly. Drawback is that it may disable ordering.
-	for _, s := range w.Structs {
-		s.Disable = true
-	}
-
 	w.Conjuncts = slices.Clone(v.Conjuncts)
 
 	for i, c := range w.Conjuncts {
@@ -905,9 +824,6 @@ func isFinal(v Value, isClosed bool) bool {
 	case *Vertex:
 		closed := isClosed || x.ClosedNonRecursive || x.ClosedRecursive
 
-		// TODO(evalv3): this is for V2 compatibility. Remove once V2 is gone.
-		closed = closed || x.IsClosedList() || x.IsClosedStruct()
-
 		// This also dereferences the value.
 		if v, ok := x.BaseValue.(Value); ok {
 			return isFinal(v, closed)
@@ -984,23 +900,19 @@ func Unify(c *OpContext, a, b Value) *Vertex {
 	// early enough to error on schemas used for validation.
 	if n := c.vertex; n != nil {
 		v.Parent = n.Parent
-		if c.isDevVersion() {
-			v.Label = n.Label // this doesn't work in V2
-		}
+		v.Label = n.Label
 	}
 
 	addConjuncts(c, v, a)
 	addConjuncts(c, v, b)
 
-	if c.isDevVersion() {
-		s := v.getState(c)
-		// As this is a new node, we should drop all the requirements from
-		// parent nodes, as these will not be aligned with the reinsertion
-		// of the conjuncts.
-		s.dropParentRequirements = true
-		if p := c.vertex; p != nil && p.state != nil && s != nil {
-			s.hasNonCyclic = p.state.hasNonCyclic
-		}
+	s := v.getState(c)
+	// As this is a new node, we should drop all the requirements from
+	// parent nodes, as these will not be aligned with the reinsertion
+	// of the conjuncts.
+	s.dropParentRequirements = true
+	if p := c.vertex; p != nil && p.state != nil && s != nil {
+		s.hasNonCyclic = p.state.hasNonCyclic
 	}
 
 	v.Finalize(c)
@@ -1018,38 +930,35 @@ func addConjuncts(ctx *OpContext, dst *Vertex, src Value) {
 	c := MakeConjunct(nil, src, closeInfo)
 
 	if v, ok := src.(*Vertex); ok {
-		if ctx.Version == internal.EvalV2 {
-			if v.ClosedRecursive {
-				var root CloseInfo
-				c.CloseInfo = root.SpawnRef(v, v.ClosedRecursive, nil)
-			}
-		} else {
-			// By default, all conjuncts in a node are considered to be not
-			// mutually closed. This means that if one of the arguments to Unify
-			// closes, but is acquired to embedding, the closeness information
-			// is disregarded. For instance, for Unify(a, b) where a and b are
-			//
-			//		a:  {#D, #D: d: f: int}
-			//		b:  {d: e: 1}
-			//
-			// we expect 'e' to be not allowed.
-			//
-			// In order to do so, we wrap the outer conjunct in a separate
-			// scope that will be closed in the presence of closed embeddings
-			// independently from the other conjuncts.
-			n := dst.getBareState(ctx)
-			c.CloseInfo = n.splitScope(c.CloseInfo)
+		// TODO(v1.0.0): we should determine whether to apply the new semantics
+		// for closedness. However, this is not applicable for a Vertex.
+		// Ultimately, this logic should be removed.
 
-			// Even if a node is marked as ClosedRecursive, it may be that this
-			// is the first node that references a definition.
-			// We approximate this to see if the path leading up to this
-			// value is a defintion. This is not fully accurate. We could
-			// investigate the closedness information contained in the parent.
-			for p := v; p != nil; p = p.Parent {
-				if p.Label.IsDef() {
-					c.CloseInfo.TopDef = true
-					break
-				}
+		// By default, all conjuncts in a node are considered to be not
+		// mutually closed. This means that if one of the arguments to Unify
+		// closes, but is acquired to embedding, the closeness information
+		// is disregarded. For instance, for Unify(a, b) where a and b are
+		//
+		//		a:  {#D, #D: d: f: int}
+		//		b:  {d: e: 1}
+		//
+		// we expect 'e' to be not allowed.
+		//
+		// In order to do so, we wrap the outer conjunct in a separate
+		// scope that will be closed in the presence of closed embeddings
+		// independently from the other conjuncts.
+		n := dst.getBareState(ctx)
+		c.CloseInfo = n.splitScope(nil, c.CloseInfo)
+
+		// Even if a node is marked as ClosedRecursive, it may be that this
+		// is the first node that references a definition.
+		// We approximate this to see if the path leading up to this
+		// value is a defintion. This is not fully accurate. We could
+		// investigate the closedness information contained in the parent.
+		for p := v; p != nil; p = p.Parent {
+			if p.Label.IsDef() {
+				c.CloseInfo.TopDef = true
+				break
 			}
 		}
 	}
@@ -1062,28 +971,40 @@ func (v *Vertex) Finalize(c *OpContext) {
 	// case the caller did not handle existing errors in the context.
 	err := c.errs
 	c.errs = nil
-	c.unify(v, combinedFlags{
-		status:    finalized,
-		condition: allKnown,
-		mode:      finalize,
+	c.unify(v, Flags{
+		status:     finalized,
+		condition:  allKnown,
+		mode:       finalize,
+		checkTypos: true,
 	})
 	c.errs = err
 }
 
+func (v *Vertex) Unify(c *OpContext, flags Flags) {
+	// Saving and restoring the error context prevents v from panicking in
+	// case the caller did not handle existing errors in the context.
+	err := c.errs
+	c.errs = nil
+	c.unify(v, flags)
+	c.errs = err
+}
+
 // CompleteArcs ensures the set of arcs has been computed.
 func (v *Vertex) CompleteArcs(c *OpContext) {
-	c.unify(v, combinedFlags{
-		status:    conjuncts,
-		condition: allKnown,
-		mode:      finalize,
+	c.unify(v, Flags{
+		status:     conjuncts,
+		condition:  allKnown,
+		mode:       finalize,
+		checkTypos: true,
 	})
 }
 
 func (v *Vertex) CompleteArcsOnly(c *OpContext) {
-	c.unify(v, combinedFlags{
-		status:    conjuncts,
-		condition: fieldSetKnown,
-		mode:      finalize,
+	c.unify(v, Flags{
+		status:     conjuncts,
+		condition:  fieldSetKnown,
+		mode:       finalize,
+		checkTypos: false,
 	})
 }
 
@@ -1153,19 +1074,6 @@ func Unwrap(v Value) Value {
 	return x.Value()
 }
 
-// OptionalType is a bit field of the type of optional constraints in use by an
-// Acceptor.
-type OptionalType int8
-
-const (
-	HasField          OptionalType = 1 << iota // X: T
-	HasDynamic                                 // (X): T or "\(X)": T
-	HasPattern                                 // [X]: T
-	HasComplexPattern                          // anything but a basic type
-	HasAdditional                              // ...T
-	IsOpen                                     // Defined for all fields
-)
-
 func (v *Vertex) Kind() Kind {
 	// This is possible when evaluating comprehensions. It is potentially
 	// not known at this time what the type is.
@@ -1181,14 +1089,6 @@ func (v *Vertex) Kind() Kind {
 	}
 }
 
-func (v *Vertex) OptionalTypes() OptionalType {
-	var mask OptionalType
-	for _, s := range v.Structs {
-		mask |= s.OptionalTypes()
-	}
-	return mask
-}
-
 // IsOptional reports whether a field is explicitly defined as optional,
 // as opposed to whether it is allowed by a pattern constraint.
 func (v *Vertex) IsOptional(label Feature) bool {
@@ -1204,17 +1104,33 @@ func (v *Vertex) accepts(ok, required bool) bool {
 	return ok || (!required && !v.ClosedRecursive)
 }
 
+// IsOpenStruct reports whether any field that is not contained within v is allowed.
+//
+// TODO: merge this function with IsClosedStruct and possibly IsClosedList.
+// right now this causes too many issues if we do so.
+func (v *Vertex) IsOpenStruct() bool {
+	// TODO: move this check to IsClosedStruct. Right now this causes too many
+	// changes in the debug output, and it also appears to be not entirely
+	// correct.
+	if v.HasEllipsis {
+		return true
+	}
+	if v.ClosedNonRecursive {
+		return false
+	}
+	if v.IsClosedStruct() {
+		return false
+	}
+	return true
+}
+
 func (v *Vertex) IsClosedStruct() bool {
-	// TODO: uncomment this. This fixes a bunch of closedness bugs
-	// in the old and new evaluator. For compability sake, though, we
-	// keep it as is for now.
-	// if v.Closed {
-	// 	return true
-	// }
+	// TODO: add this check. Right now this causes issues. It will have
+	// to be carefully introduced.
 	// if v.HasEllipsis {
 	// 	return false
 	// }
-	switch x := v.BaseValue.(type) {
+	switch v.BaseValue.(type) {
 	default:
 		return false
 
@@ -1222,10 +1138,6 @@ func (v *Vertex) IsClosedStruct() bool {
 		return v.ClosedRecursive && !v.HasEllipsis
 
 	case *StructMarker:
-		if x.NeedClose {
-			return true
-		}
-
 	case *Disjunction:
 	}
 	return isClosed(v)
@@ -1270,7 +1182,7 @@ func (v *Vertex) Accept(ctx *OpContext, f Feature) bool {
 		switch v.BaseValue.(type) {
 		case *ListMarker:
 			// TODO(perf): use precomputed length.
-			if f.Index() < len(v.Elems()) {
+			if f.Index() < iterutil.Count(v.Elems()) {
 				return true
 			}
 			return !v.IsClosedList()
@@ -1288,15 +1200,7 @@ func (v *Vertex) Accept(ctx *OpContext, f Feature) bool {
 		}
 	}
 
-	// TODO: move this check to IsClosedStruct. Right now this causes too many
-	// changes in the debug output, and it also appears to be not entirely
-	// correct.
-	if v.HasEllipsis {
-		return true
-
-	}
-
-	if !v.IsClosedStruct() || v.Lookup(f) != nil {
+	if v.IsOpenStruct() || v.Lookup(f) != nil {
 		return true
 	}
 
@@ -1315,15 +1219,6 @@ func (v *Vertex) MatchAndInsert(ctx *OpContext, arc *Vertex) {
 	}
 
 	// Go backwards to simulate old implementation.
-	if !ctx.isDevVersion() {
-		for i := len(v.Structs) - 1; i >= 0; i-- {
-			s := v.Structs[i]
-			if s.Disable {
-				continue
-			}
-			s.MatchAndInsert(ctx, arc)
-		}
-	}
 
 	// This is the equivalent for the new implementation.
 	if pcs := v.PatternConstraints; pcs != nil {
@@ -1387,15 +1282,17 @@ func (v *Vertex) LookupRaw(f Feature) *Vertex {
 }
 
 // Elems returns the regular elements of a list.
-func (v *Vertex) Elems() []*Vertex {
-	// TODO: add bookkeeping for where list arcs start and end.
-	a := make([]*Vertex, 0, len(v.Arcs))
-	for _, x := range v.Arcs {
-		if x.Label.IsInt() {
-			a = append(a, x)
+func (v *Vertex) Elems() iter.Seq[*Vertex] {
+	return func(yield func(*Vertex) bool) {
+		// TODO: add bookkeeping for where list arcs start and end.
+		for _, x := range v.Arcs {
+			if x.Label.IsInt() {
+				if !yield(x) {
+					break
+				}
+			}
 		}
 	}
-	return a
 }
 
 func (v *Vertex) Init(c *OpContext) {
@@ -1411,33 +1308,7 @@ func (v *Vertex) GetArc(c *OpContext, f Feature, t ArcType) (arc *Vertex, isNew
 		return arc, false
 	}
 
-	if c.isDevVersion() {
-		return nil, false
-	}
-
-	if v.LockArcs {
-		// TODO(errors): add positions.
-		if f.IsInt() {
-			c.addErrf(EvalError, token.NoPos,
-				"element at index %v not allowed by earlier comprehension or reference cycle", f)
-		} else {
-			c.addErrf(EvalError, token.NoPos,
-				"field %v not allowed by earlier comprehension or reference cycle", f)
-		}
-	}
-	// TODO: consider setting Dynamic here from parent.
-	arc = &Vertex{
-		Parent:    v,
-		Label:     f,
-		ArcType:   t,
-		nonRooted: v.IsDynamic || v.Label.IsLet() || v.nonRooted,
-		anonymous: v.anonymous || v.Label.IsLet(),
-	}
-	v.Arcs = append(v.Arcs, arc)
-	if t == ArcPending {
-		v.hasPendingArc = true
-	}
-	return arc, true
+	return nil, false
 }
 
 func (v *Vertex) Source() ast.Node {
@@ -1502,8 +1373,7 @@ func (v *Vertex) hasConjunct(c Conjunct) (added bool) {
 func findConjunct(cs []Conjunct, c Conjunct) (int, Conjunct) {
 	for i, x := range cs {
 		// TODO: disregard certain fields from comparison (e.g. Refs)?
-		if x.CloseInfo.closeInfo == c.CloseInfo.closeInfo && // V2
-			x.x == c.x &&
+		if x.x == c.x &&
 			x.Env.Up == c.Env.Up && x.Env.Vertex == c.Env.Vertex {
 			return i, x
 		}
@@ -1511,67 +1381,8 @@ func findConjunct(cs []Conjunct, c Conjunct) (int, Conjunct) {
 	return -1, Conjunct{}
 }
 
-func (n *nodeContext) addConjunction(c Conjunct, index int) {
-	unreachableForDev(n.ctx)
-
-	// NOTE: This does not split binary expressions for comprehensions.
-	// TODO: split for comprehensions and rewrap?
-	if x, ok := c.Elem().(*BinaryExpr); ok && x.Op == AndOp {
-		c.x = x.X
-		n.conjuncts = append(n.conjuncts, conjunct{C: c, index: index})
-		c.x = x.Y
-		n.conjuncts = append(n.conjuncts, conjunct{C: c, index: index})
-	} else {
-		n.conjuncts = append(n.conjuncts, conjunct{C: c, index: index})
-	}
-}
-
 func (v *Vertex) addConjunctUnchecked(c Conjunct) {
-	index := len(v.Conjuncts)
 	v.Conjuncts = append(v.Conjuncts, c)
-	if n := v.state; n != nil && !n.ctx.isDevVersion() {
-		// TODO(notify): consider this as a central place to send out
-		// notifications. At the moment this is not necessary, but it may
-		// be if we move the notification mechanism outside of the path of
-		// running tasks.
-		n.addConjunction(c, index)
-
-		// TODO: can we remove notifyConjunct here? This method is only
-		// used if either Unprocessed is 0, in which case there will be no
-		// notification recipients, or for "pushed down" comprehensions,
-		// which should also have been added at an earlier point.
-		n.notifyConjunct(c)
-	}
-}
-
-// addConjunctDynamic adds a conjunct to a vertex and immediately evaluates
-// it, whilst doing the same for any vertices on the notify list, recursively.
-func (n *nodeContext) addConjunctDynamic(c Conjunct) {
-	unreachableForDev(n.ctx)
-
-	n.node.Conjuncts = append(n.node.Conjuncts, c)
-	n.addExprConjunct(c, partial)
-	n.notifyConjunct(c)
-
-}
-
-func (n *nodeContext) notifyConjunct(c Conjunct) {
-	unreachableForDev(n.ctx)
-
-	for _, rec := range n.notify {
-		arc := rec.v
-		if !arc.hasConjunct(c) {
-			if arc.state == nil {
-				// TODO: continuing here is likely to result in a faulty
-				// (incomplete) configuration. But this may be okay. The
-				// CUE_DEBUG=0 flag disables this assertion.
-				n.ctx.Assertf(n.ctx.pos(), !n.ctx.Strict, "unexpected nil state")
-				n.ctx.addErrf(0, n.ctx.pos(), "cannot add to field %v", arc.Label)
-				continue
-			}
-			arc.state.addConjunctDynamic(c)
-		}
-	}
 }
 
 func (v *Vertex) AddStruct(s *StructLit, env *Environment, ci CloseInfo) *StructInfo {
