@@ -582,10 +582,6 @@ type CycleInfo struct {
 	// TODO: make this a method and use CycleType == IsCyclic after V2 is removed.
 	IsCyclic bool
 
-	// Inline is used to detect expressions referencing themselves, for instance:
-	//     {x: out, out: x}.out
-	Inline bool
-
 	// TODO(perf): pack this in with CloseInfo. Make an uint32 pointing into
 	// a buffer maintained in OpContext, using a mark-release mechanism.
 	Refs *RefNode
@@ -627,7 +623,7 @@ type cyclicConjunct struct {
 	arc *Vertex // cached Vertex
 }
 
-// CycleType indicates the type of cycle detected. The CyclicType is associated
+// CyclicType indicates the type of cycle detected. The CyclicType is associated
 // with a conjunct and may only increase in value for child conjuncts.
 type CyclicType uint8
 
@@ -819,327 +815,10 @@ func (c *CloseInfo) setOptionalV3(n *nodeContext) {
 	}
 }
 
-// markCycle checks whether the reference x is cyclic. There are two cases:
-//  1. it was previously used in this conjunct, and
-//  2. it directly references a parent node.
-//
-// Other inputs:
-//
-//	arc      the reference to which x points
-//	env, ci  the components of the Conjunct from which x originates
-//
-// A cyclic node is added to a queue for later processing if no evidence of a
-// non-cyclic node has so far been found. updateCyclicStatus processes delayed
-// nodes down the line once such evidence is found.
-//
-// If a cycle is the result of "inline" processing (an expression referencing
-// itself), an error is reported immediately.
-//
-// It returns the CloseInfo with tracked cyclic conjuncts updated, and
-// whether or not its processing should be skipped, which is the case either if
-// the conjunct seems to be fully cyclic so far or if there is a valid reference
-// cycle.
-func (n *nodeContext) markCycle(arc *Vertex, env *Environment, x Resolver, ci CloseInfo) (_ CloseInfo, skip bool) {
-	unreachableForDev(n.ctx)
-
-	n.assertInitialized()
-
-	// TODO(perf): this optimization can work if we also check for any
-	// references pointing to arc within arc. This can be done with compiler
-	// support. With this optimization, almost all references could avoid cycle
-	// checking altogether!
-	// if arc.status == Finalized && arc.cyclicReferences == nil {
-	//  return v, false
-	// }
-
-	// Check whether the reference already occurred in the list, signaling
-	// a potential cycle.
-	found := false
-	depth := int32(0)
-	for r := ci.Refs; r != nil; r = r.Next {
-		if r.Ref != x {
-			// TODO(share): this is a bit of a hack. We really should implement
-			// (*Vertex).cyclicReferences for the new evaluator. However,
-			// implementing cyclicReferences is somewhat tricky, as it requires
-			// referenced nodes to be evaluated, which is a guarantee we may not
-			// want to give. Moreover, it seems we can find a simpler solution
-			// based on structure sharing. So punt on this solution for now.
-			if r.Arc != arc || !n.ctx.isDevVersion() {
-				continue
-			}
-			found = true
-		}
-
-		// A reference that is within a graph that is being evaluated
-		// may repeat with a different arc and will point to a
-		// non-finalized arc. A repeating reference that points outside the
-		// graph will always be the same address. Hence, if this is a
-		// finalized arc with a different address, it resembles a reference that
-		// is included through a different path and is not a cycle.
-		if !equalDeref(r.Arc, arc) && arc.status == finalized {
-			continue
-		}
-
-		// For dynamically created structs we mark this as an error. Otherwise
-		// there is only an error if we have visited the arc before.
-		if ci.Inline && (arc.IsDynamic || equalDeref(r.Arc, arc)) {
-			n.reportCycleError()
-			return ci, true
-		}
-
-		// We have a reference cycle, as distinguished from a structural
-		// cycle. Reference cycles represent equality, and thus are equal
-		// to top. We can stop processing here.
-		// var nn1, nn2 *Vertex
-		// if u := r.Node.state.underlay; u != nil {
-		// 	nn1 = u.node
-		// }
-		// if u := n.node.state.underlay; u != nil {
-		// 	nn2 = u.node
-		// }
-		if equalDeref(r.Node, n.node) {
-			return ci, true
-		}
-
-		depth = r.Depth
-		found = true
-
-		// Mark all conjuncts of this Vertex that refer to the same node as
-		// cyclic. This is an extra safety measure to ensure that two conjuncts
-		// cannot work in tandom to circumvent a cycle. It also tightens
-		// structural cycle detection in some cases. Late detection of cycles
-		// can result in a lot of redundant work.
-		//
-		// TODO: this loop is not on a critical path, but it may be evaluated
-		// if it is worthy keeping at some point.
-		for i, c := range n.node.Conjuncts {
-			if c.CloseInfo.IsCyclic {
-				continue
-			}
-			for rr := c.CloseInfo.Refs; rr != nil; rr = rr.Next {
-				// TODO: Is it necessary to find another way to find
-				// "parent" conjuncts? This mechanism seems not entirely
-				// accurate. Maybe a pointer up to find the root and then
-				// "spread" downwards?
-				if r.Ref == x && equalDeref(r.Arc, rr.Arc) {
-					n.node.Conjuncts[i].CloseInfo.IsCyclic = true
-					break
-				}
-			}
-		}
-
-		break
-	}
-
-	if arc.state != nil {
-		if d := arc.state.evalDepth; d > 0 && d >= n.ctx.optionalMark {
-			arc.IsCyclic = true
-		}
-	}
-
-	// The code in this switch statement registers structural cycles caught
-	// through EvaluatingArcs to the root of the cycle. This way, any node
-	// referencing this value can track these nodes early. This is mostly an
-	// optimization to shorten the path for which structural cycles are
-	// detected, which may be critical for performance.
-outer:
-	switch arc.status {
-	case evaluatingArcs: // also  Evaluating?
-		if arc.state.evalDepth < n.ctx.optionalMark {
-			break
-		}
-
-		// The reference may already be there if we had no-cyclic structure
-		// invalidating the cycle.
-		for r := arc.cyclicReferences; r != nil; r = r.Next {
-			if r.Ref == x {
-				break outer
-			}
-		}
-
-		arc.cyclicReferences = &RefNode{
-			Arc:  deref(arc),
-			Ref:  x,
-			Next: arc.cyclicReferences,
-		}
-
-	case finalized:
-		// Insert cyclic references from found arc, if any.
-		for r := arc.cyclicReferences; r != nil; r = r.Next {
-			if r.Ref == x {
-				// We have detected a cycle, with the only exception if arc is
-				// a disjunction, as evaluation always stops at unresolved
-				// disjunctions.
-				if _, ok := arc.BaseValue.(*Disjunction); !ok {
-					found = true
-				}
-			}
-			ci.Refs = &RefNode{
-				Arc:  deref(r.Arc),
-				Node: deref(n.node),
-
-				Ref:   x,
-				Next:  ci.Refs,
-				Depth: n.depth,
-			}
-		}
-	}
-
-	// NOTE: we need to add a tracked reference even if arc is not cyclic: it
-	// may still cause a cycle that does not refer to a parent node. For
-	// instance:
-	//
-	//      y: [string]: b: y
-	//      x: y
-	//      x: c: x
-	//
-	// ->
-	//          - in conjuncts
-	//             - out conjuncts: these count for cycle detection.
-	//      x: {
-	//          [string]: <1: y> b: y
-	//          c: x
-	//      }
-	//      x.c: {
-	//          <1: y> b: y
-	//          <2: x> y
-	//             [string]: <3: x, y> b: y
-	//          <2: x> c: x
-	//      }
-	//      x.c.b: {
-	//          <1: y> y
-	//             [string]: <4: y; Cyclic> b: y
-	//          <3: x, y> b: y
-	//      }
-	//      x.c.b.b: {
-	//          <3: x, y> y
-	//               [string]: <5: x, y, Cyclic> b: y
-	//          <4: y, Cyclic> y
-	//               [string]: <5: x, y, Cyclic> b: y
-	//      }
-	//      x.c.c: { // structural cycle
-	//          <3: x, y> b: y
-	//          <2: x> x
-	//               <6: x, Cyclic>: y
-	//                    [string]: <8: x, y; Cyclic> b: y
-	//               <7: x, Cyclic>: c: x
-	//      }
-	//      x.c.c.b: { // structural cycle
-	//          <3: x, y> y
-	//               [string]: <3: x, y; Cyclic> b: y
-	//          <8: x, y; Cyclic> y
-	//      }
-	// ->
-	//      x: [string]: b: y
-	//      x: c: b: y
-	//      x: c: [string]: b: y
-	//      x: c: b: b: y
-	//      x: c: b: [string]: b: y
-	//      x: c: b: b: b: y
-	//      ....       // structural cycle 1
-	//      x: c: c: x // structural cycle 2
-	//
-	// Note that in this example there is a structural cycle at x.c.c, but we
-	// would need go guarantee that cycle is detected before the algorithm
-	// descends into x.c.b.
-	if !found || depth != n.depth {
-		// Adding this in case there is a definite cycle is unnecessary, but
-		// gives somewhat better error messages.
-		// We also need to add the reference again if the depth differs, as
-		// the depth is used for tracking "new structure".
-		// var nn *Vertex
-		// if u := n.node.state.underlay; u != nil {
-		// 	nn = u.node
-		// }
-		ci.Refs = &RefNode{
-			Arc:   deref(arc),
-			Ref:   x,
-			Node:  deref(n.node),
-			Next:  ci.Refs,
-			Depth: n.depth,
-		}
-	}
-
-	if !found && arc.status != evaluatingArcs {
-		// No cycle.
-		return ci, false
-	}
-
-	// TODO: consider if we should bail if a cycle is detected using this
-	// mechanism. Ultimately, especially when the old evaluator is removed
-	// and the status field purged, this should be used instead of the above.
-	// if !found && arc.state.evalDepth < n.ctx.optionalMark {
-	// 	// No cycle.
-	// 	return ci, false
-	// }
-
-	alreadyCycle := ci.IsCyclic
-	ci.IsCyclic = true
-
-	// TODO: depth might legitimately be 0 if it is a root vertex.
-	// In the worst case, this may lead to a spurious cycle.
-	// Fix this by ensuring the root vertex starts with a depth of 1, for
-	// instance.
-	if depth > 0 {
-		// Look for evidence of "new structure" to invalidate the cycle.
-		// This is done by checking for non-cyclic conjuncts between the
-		// current vertex up to the ancestor to which the reference points.
-		// Note that the cyclic conjunct may not be marked as such, so we
-		// look for at least one other non-cyclic conjunct if this is the case.
-		upCount := n.depth - depth
-		for p := n.node.Parent; p != nil; p = p.Parent {
-			if upCount--; upCount <= 0 {
-				break
-			}
-			a := p.Conjuncts
-			count := 0
-			for _, c := range a {
-				count += getNonCyclicCount(c)
-			}
-			if !alreadyCycle {
-				count--
-			}
-			if count > 0 {
-				return ci, false
-			}
-		}
-	}
-
-	n.hasAnyCyclicConjunct = true
-	if !n.hasNonCycle && env != nil {
-		// TODO: investigate if we can get rid of cyclicConjuncts in the new
-		// evaluator.
-		v := Conjunct{env, x, ci}
-		n.cyclicConjuncts = append(n.cyclicConjuncts, cyclicConjunct{v, arc})
-		return ci, true
-	}
-
-	return ci, false
-}
-
-func getNonCyclicCount(c Conjunct) int {
-	switch a, ok := c.x.(*ConjunctGroup); {
-	case ok:
-		count := 0
-		for _, c := range *a {
-			count += getNonCyclicCount(c)
-		}
-		return count
-
-	case !c.CloseInfo.IsCyclic:
-		return 1
-
-	default:
-		return 0
-	}
-}
-
 // updateCyclicStatusV3 looks for proof of non-cyclic conjuncts to override
 // a structural cycle.
 func (n *nodeContext) updateCyclicStatusV3(c CloseInfo) {
-	if n.ctx.inDisjunct == 0 {
-		n.hasFieldValue = true
-	}
+	n.hasFieldValue = true
 	if !c.IsCyclic {
 		n.hasNonCycle = true
 		for _, c := range n.cyclicConjuncts {
@@ -1154,20 +833,6 @@ func (n *nodeContext) updateCyclicStatusV3(c CloseInfo) {
 	}
 }
 
-// updateCyclicStatus looks for proof of non-cyclic conjuncts to override
-// a structural cycle.
-func (n *nodeContext) updateCyclicStatus(c CloseInfo) {
-	unreachableForDev(n.ctx)
-
-	if !c.IsCyclic {
-		n.hasNonCycle = true
-		for _, c := range n.cyclicConjuncts {
-			n.addVertexConjuncts(c.c, c.arc, false)
-		}
-		n.cyclicConjuncts = n.cyclicConjuncts[:0]
-	}
-}
-
 func assertStructuralCycleV3(n *nodeContext) bool {
 	n.cyclicConjuncts = n.cyclicConjuncts[:0]
 
@@ -1178,14 +843,6 @@ func assertStructuralCycleV3(n *nodeContext) bool {
 	return false
 }
 
-func assertStructuralCycle(n *nodeContext) bool {
-	if n.hasAnyCyclicConjunct && !n.hasNonCycle {
-		n.reportCycleError()
-		return true
-	}
-	return false
-}
-
 func (n *nodeContext) reportCycleError() {
 	b := &Bottom{
 		Code:  StructuralCycleError,
@@ -1209,8 +866,7 @@ func (n *nodeContext) reportCycleError() {
 func makeAnonymousConjunct(env *Environment, x Expr, refs *RefNode) Conjunct {
 	return Conjunct{
 		env, x, CloseInfo{CycleInfo: CycleInfo{
-			Inline: true,
-			Refs:   refs,
+			Refs: refs,
 		}},
 	}
 }
@@ -1227,35 +883,6 @@ func (n *nodeContext) decDepth() {
 	n.ctx.evalDepth--
 }
 
-// markOptional marks that we are about to process an "optional element" that
-// allows errors. In these cases, structural cycles are not "terminal".
-//
-// Examples of such constructs are:
-//
-// Optional fields:
-//
-//	a: b?: a
-//
-// Pattern constraints:
-//
-//	a: [string]: a
-//
-// Disjunctions:
-//
-//	a: b: null | a
-//
-// A call to markOptional should be paired with a call to unmarkOptional.
-func (n *nodeContext) markOptional() (saved int) {
-	saved = n.ctx.evalDepth
-	n.ctx.optionalMark = n.ctx.evalDepth
-	return saved
-}
-
-// See markOptional.
-func (n *nodeContext) unmarkOptional(saved int) {
-	n.ctx.optionalMark = saved
-}
-
 // markDepth assigns the current evaluation depth to the receiving node.
 // Any previously assigned depth is saved and returned and should be restored
 // using unmarkDepth after processing n.
