@@ -166,7 +166,6 @@ package adt
 // type.
 //
 import (
-	"math"
 	"slices"
 
 	"cuelang.org/go/cue/ast"
@@ -174,41 +173,33 @@ import (
 
 type defID uint32
 
+//go:generate go tool stringer -type=defIDType -linecomment
+
 type defIDType int8
 
 const (
 	// defIDTypeUnknown indicates that the ID is not a definition.
-	defIDTypeUnknown defIDType = iota
+	defIDTypeUnknown defIDType = iota // *
 
-	defEmbedding
-	defReference
-	defStruct
+	defEmbedding // E
+	defReference // D
+	defStruct    // S
 )
 
-func (d defIDType) String() string {
-	switch d {
-	case defEmbedding:
-		return "E"
-	case defReference:
-		return "D"
-	case defStruct:
-		return "S"
-	default:
-		return "*"
-	}
+type containment struct {
+	id defID
+	n  Node
 }
 
-const deleteID defID = math.MaxUint32
-
-func (c *OpContext) getNextDefID() defID {
+func (c *OpContext) getNextDefID(n Node) defID {
 	c.stats.NumCloseIDs++
 	c.nextDefID++
 
 	if len(c.containments) == 0 {
 		// Our ID starts at 1. Create an extra element for the zero value.
-		c.containments = make([]defID, 1, 16)
+		c.containments = make([]containment, 1, 16)
 	}
-	c.containments = append(c.containments, 0)
+	c.containments = append(c.containments, containment{id: 0, n: n})
 
 	return c.nextDefID
 }
@@ -275,8 +266,8 @@ func (n *nodeContext) addReplacement(x replaceID) {
 		return
 	}
 
-	if x.from < x.to && n.ctx.containments[x.to] == 0 {
-		n.ctx.containments[x.to] = x.from
+	if x.from < x.to && n.ctx.containments[x.to].id == 0 {
+		n.ctx.containments[x.to].id = x.from
 		return
 	}
 
@@ -320,7 +311,7 @@ func (n *nodeContext) updateConjunctInfo(k Kind, id CloseInfo, flags conjunctFla
 // multiple times to a single node. As we only want to insert each conjunct
 // once, we need to ensure that within all contexts a single ID assigned to such
 // a resolver is tracked.
-func (n *nodeContext) addResolver(v *Vertex, id CloseInfo, forceIgnore bool) CloseInfo {
+func (n *nodeContext) addResolver(p Node, v *Vertex, id CloseInfo, forceIgnore bool) CloseInfo {
 	if n.ctx.OpenDef {
 		return id
 	}
@@ -329,7 +320,8 @@ func (n *nodeContext) addResolver(v *Vertex, id CloseInfo, forceIgnore bool) Clo
 		return id
 	}
 
-	closeOuter := (id.FromDef && id.FromEmbed) || v.ClosedNonRecursive
+	// This can only be true when not using the @experiment(explicitopen).
+	closeOuter := id.FromDef && id.FromEmbed && !id.Opened
 
 	if closeOuter && !forceIgnore {
 		// Walk up the parent chain of the outer structs to "activate" them.
@@ -348,7 +340,7 @@ func (n *nodeContext) addResolver(v *Vertex, id CloseInfo, forceIgnore bool) Clo
 
 	var ignore bool
 	switch {
-	case forceIgnore:
+	case forceIgnore, id.Opened:
 		// Special mode to always ignore the outer enclosing group.
 		// This is the case, for instance, if a resolver resolves to a
 		// non-definition.
@@ -366,7 +358,7 @@ func (n *nodeContext) addResolver(v *Vertex, id CloseInfo, forceIgnore bool) Clo
 		// 			// the embedding.
 		// 			#A & #B
 		// 		}
-		isClosed := id.FromDef || v.ClosedNonRecursive
+		isClosed := id.FromDef || v.ClosedNonRecursive || v.ClosedRecursive
 		ignore = !isClosed
 	default:
 		// In the default case we can disable typo checking this type if it is
@@ -375,15 +367,21 @@ func (n *nodeContext) addResolver(v *Vertex, id CloseInfo, forceIgnore bool) Clo
 	}
 
 	dstID := defID(0)
-	for _, x := range n.reqDefIDs {
+	for i, x := range n.reqDefIDs {
 		if x.v == v {
 			dstID = x.id
+			if x.ignore && !ignore {
+				// Override settings of a vertex added by #A...
+				n.reqDefIDs[i].ignore = false
+				n.reqDefIDs[i].parent = id.outerID
+				n.reqDefIDs[i].embed = id.enclosingEmbed
+			}
 			break
 		}
 	}
 
 	if dstID == 0 || id.enclosingEmbed != 0 {
-		next := n.ctx.getNextDefID()
+		next := n.ctx.getNextDefID(p)
 		if dstID != 0 {
 			// If we need to activate an enclosing embed group, and the added
 			// resolver was already before, we need to allocate a new ID and
@@ -418,6 +416,10 @@ func (c *OpContext) subField(ci CloseInfo) CloseInfo {
 	// as, at this point, it seems to be only used for debugging. We may
 	// want to consider having a separate field for this, though.
 	ci.FromEmbed = false
+
+	// The priority should be cleared for sub fields: we take default struct
+	// in its entirety, but not piecemeal.
+	ci.Priority = 0
 	return ci
 }
 
@@ -431,12 +433,12 @@ func (id CloseInfo) clearCloseCheck() CloseInfo {
 	return id
 }
 
-func (n *nodeContext) newReq(id CloseInfo, kind defIDType) CloseInfo {
+func (n *nodeContext) newReq(p Node, id CloseInfo, kind defIDType) CloseInfo {
 	if id.defID != 0 && id.opID != n.ctx.opID {
 		return id.clearCloseCheck()
 	}
 
-	dstID := n.ctx.getNextDefID()
+	dstID := n.ctx.getNextDefID(p)
 	n.addReplacement(replaceID{from: id.defID, to: dstID})
 
 	parent := id.defID
@@ -496,6 +498,10 @@ func (v *Vertex) AddOpenConjunct(ctx *OpContext, w *Vertex) {
 // We can then say that requirement 3 (node A) holds if all fields contain
 // either label 3, or any field within 1 that is not 2.
 func (n *nodeContext) injectEmbedNode(x Decl, id CloseInfo) CloseInfo {
+	if pos(x).Experiment().ExplicitOpen {
+		return id
+	}
+
 	id.FromEmbed = true
 
 	// Filter cases where we do not need to track the definition.
@@ -506,7 +512,7 @@ func (n *nodeContext) injectEmbedNode(x Decl, id CloseInfo) CloseInfo {
 		}
 	}
 
-	return n.newReq(id, defEmbedding)
+	return n.newReq(x, id, defEmbedding)
 }
 
 // splitStruct is used to mark the outer struct of a field in which embeddings
@@ -520,7 +526,7 @@ func (n *nodeContext) injectEmbedNode(x Decl, id CloseInfo) CloseInfo {
 // definition is embedded within a struct. It can be removed if we implement
 // the #A vs #A... semantics.
 func (n *nodeContext) splitStruct(s *StructLit, id CloseInfo) CloseInfo {
-	if n.ctx.OpenDef {
+	if pos(s).Experiment().ExplicitOpen || n.ctx.OpenDef {
 		return id
 	}
 
@@ -546,11 +552,11 @@ func (n *nodeContext) splitStruct(s *StructLit, id CloseInfo) CloseInfo {
 		return id
 	}
 
-	return n.splitScope(id)
+	return n.splitScope(s, id)
 }
 
-func (n *nodeContext) splitScope(id CloseInfo) CloseInfo {
-	return n.newReq(id, defStruct)
+func (n *nodeContext) splitScope(p Node, id CloseInfo) CloseInfo {
+	return n.newReq(p, id, defStruct)
 }
 
 func (n *nodeContext) checkTypos() {
@@ -623,7 +629,9 @@ func (n *nodeContext) checkTypos() {
 
 		// TODO: do not descend on optional?
 
-		// openDebugGraph(ctx, a, "NOT ALLOWED") // Uncomment for debugging.
+		if OpenGraphs {
+			openDebugGraph(ctx, a, "NOT ALLOWED") // Uncomment for debugging.
+		}
 
 		if b := ctx.notAllowedError(a); b != nil && a.ArcType <= ArcRequired {
 			err = CombineErrors(nil, err, b)
@@ -668,17 +676,12 @@ func (n *nodeContext) hasEvidenceForOne(all reqSets, i uint32, conjuncts []conju
 	}
 
 	embedScope, ok := all.lookupSet(a.embed)
-
 	if !ok {
 		return false
 	}
 
 	outerScope, ok := all.lookupSet(a.parent)
 
-	if ok && outerScope.removed {
-		return true
-	}
-
 outer:
 	for _, c := range conjuncts {
 		if n.containsDefID(embedScope.id, c.embed) {
@@ -689,6 +692,9 @@ outer:
 		if !ok || a.parent == 0 {
 			return true
 		}
+		if outerScope.removed {
+			return true
+		}
 
 		// If this conjunct is within the outer struct, but outside the
 		// embedding scope, this means it was "added" and we do not have
@@ -701,10 +707,15 @@ outer:
 }
 
 func (n *nodeContext) containsDefID(node, child defID) bool {
-	// TODO(perf): cache result
 	// TODO(perf): we could keep track of the minimum defID that could map so
 	// that we can use this to bail out early.
 	c := n.ctx
+
+	key := [2]defID{node, child}
+	if result, ok := n.containsDefIDCache[key]; ok {
+		return result
+	}
+
 	c.redirectsBuf = c.redirectsBuf[:0]
 	for p := n; p != nil; p = p.node.Parent.state {
 		if p.opID != n.opID {
@@ -720,7 +731,19 @@ func (n *nodeContext) containsDefID(node, child defID) bool {
 		c.stats.MaxRedirect = int64(len(c.redirectsBuf))
 	}
 
-	return n.containsDefIDRec(node, child, child)
+	result := n.containsDefIDRec(node, child, child)
+
+	// Caching in [nodeContext.containsDefIDCache] adds overhead;
+	// only do it if we estimate that [nodeContext.containsDefIDRec]
+	// is doing significant work by looking at the number of replaceIDs.
+	if len(c.redirectsBuf) > 15 {
+		if n.containsDefIDCache == nil {
+			n.containsDefIDCache = make(map[[2]defID]bool)
+		}
+		n.containsDefIDCache[key] = result
+	}
+
+	return result
 }
 
 func (n *nodeContext) containsDefIDRec(node, child, start defID) bool {
@@ -744,7 +767,7 @@ func (n *nodeContext) containsDefIDRec(node, child, start defID) bool {
 			}
 		}
 
-		p = c.containments[p]
+		p = c.containments[p].id
 		if p == start {
 			// We won't match node we haven't already after one cycle.
 			return false
@@ -805,7 +828,7 @@ func mergeCloseInfo(nv, nw *nodeContext) {
 	if w == nil {
 		return
 	}
-	// Merge missing closeInfos
+	// Merge missing conjunct infos
 outer:
 	for _, wci := range nw.conjunctInfo {
 		for _, vci := range nv.conjunctInfo {
@@ -852,7 +875,7 @@ func getReqSets(n *nodeContext) reqSets {
 		return n.reqSets
 	}
 
-	a := n.reqSets
+	a := n.reqSets[:0]
 	v := n.node
 
 	if p := v.Parent; p != nil && !n.dropParentRequirements {
