@@ -8,9 +8,9 @@ import (
 
 func TestInit(t *testing.T) {
 	// This is just a smoke test to make sure it's all wired up OK.
-	t.Setenv("CUE_EXPERIMENT", "evalv3=0,cmdreferencepkg")
+	t.Setenv("CUE_EXPERIMENT", "cmdreferencepkg=0,keepvalidators")
 	err := Init()
 	qt.Assert(t, qt.IsNil(err))
-	qt.Assert(t, qt.IsFalse(Flags.EvalV3))
-	qt.Assert(t, qt.IsTrue(Flags.CmdReferencePkg))
+	qt.Assert(t, qt.IsFalse(Flags.CmdReferencePkg))
+	qt.Assert(t, qt.IsTrue(Flags.KeepValidators))
 }
